//==========================================================================================
// Name         : SBiTDRL.cpp
// Author       : Tom Hartmann
// Version      : 1.0
// Copyright    : The tool and the source code can freely be used and distributed for non-commercial
//					purposes under the condition that the paper "Sorting Signed Permutations by
//					Inverse Tandem Duplication Random Losses" by Hartmann et al. is cited.
// Description  : Tool solves the Sorting Problem and the distance Problem under the inverse
//					Tandem Duplication Random Loss model
// Date			: 03.08.2018
//==========================================================================================

#include <iostream>
#include <getopt.h>
#include <string>
#include <stdlib.h>
#include <vector>
#include <set>

#include "helpers.hpp"
#include "genome.hpp"
#include "io.hpp"

using namespace std;


/**
 * get program parameters
 * @param[in] argc argument count
 * @param[in] argv arguments
 * @param[out] fname filename
 */
void getoptions(int argc, char *argv[], string &fname, bool &print_distance);


/**
 * print usage info and exit
 */
void usage();



int main(int argc, char *argv[]){
	string filename;				// input filename
	vector<genome> genomes; 		// the genomes
	bool circular=false;			// consider genomes to be circular
	vector<string> names, 			// names of the genomes
		nmap;						// names of the elements
	bool print_distance=false;

	getoptions(argc, argv, filename, print_distance);

	if(filename==""){
		cerr << "No file name given ---> exiting!" << endl;
		usage();
	}

	read_genomes(filename, genomes, names, circular, nmap, false);	// read the gene order file
	if(genomes.size() == 0){
		cout << "No genomes in the file!" << endl;
		usage();
	}

	if(print_distance){//solve distance problem

		// build distance matrix
		std::vector< std::vector<int> > distances(genomes.size(), std::vector<int>(genomes.size(),0));

		//distances are calculated
		for(unsigned int i=0; i<genomes.size(); i++){
			for(unsigned int j=0; j<genomes.size(); j++){
				set<int>::iterator it;
				// calculate the distance i -> j
				vector<genome> sigmas;
				vector<vector<set<int>>> rhos;
				bool last;
				sort_by_iTDRL(genomes[i],genomes[j], sigmas, rhos, last);
				distances[i][j]=rhos.size();
			}
		}


		// print distance matrix
		for(unsigned int i=0; i< names.size();i++){
			cout << "\t" << names[i];
		}
		cout << endl;
		for(unsigned int i=0; i<genomes.size(); i++){
			cout << names[i];
			for(unsigned int j=0; j<genomes.size(); j++){
				cout << "\t" << distances[i][j] ;
			}
			cout << endl;
		}

	}else{
		//solve sorting problem
		for(unsigned int i=0; i<genomes.size()-1; i++){
			for(unsigned int j=i+1; j< genomes.size(); j++){
				set<int>::iterator it;
				// calculate the distance & scenario i -> j
				vector<genome> sigmas;
				vector<vector<set<int>>> rhos;
				bool last;
				sort_by_iTDRL(genomes[i],genomes[j], sigmas, rhos, last);
				cout << "Distance from " << names[i] << " to " << names[j] << " is " << rhos.size() << ":" << endl;
				for( int k=sigmas.size()-1; k>=0; --k){
					cout << sigmas[k] << endl;
					if(k!=0){
						if(k==(int)sigmas.size()-1 && last){
							cout << "\t iTDRL(l,";
						}else{
							cout << "\t iTDRL(r,";
						}
						for(it=rhos[k-1][0].begin(); it!=rhos[k-1][0].end(); ++it){
							if(*it<0){
								cout << " -" << nmap[abs(*it)];
							}else{
								cout <<  " " << nmap[*it];
							}
						}
						cout << ",";
						for(it=rhos[k-1][1].begin(); it!=rhos[k-1][1].end(); ++it){
							if(*it<0){
								cout << " -" << nmap[abs(*it)];
							}else{
								cout <<  " " << nmap[*it];
							}
						}
						cout << ")" << endl;
					}
				}
				cout << endl;

				// calculate the distance & scenario j -> i
				vector<genome> sigmas2;
				vector<vector<set<int>>> rhos2;
				bool last2;
				sort_by_iTDRL(genomes[j],genomes[i], sigmas2, rhos2, last2);
				cout << "Distance from " << names[j] << " to " << names[i] << " is " << rhos2.size() << ":" << endl;
				for(int k=sigmas2.size()-1; k>=0; k--){
					cout << sigmas2[k] << endl;
					if(k!=0){
						if(k==(int)sigmas2.size()-1 && last){
							cout << "\t iTDRL(l,";
						}else{
							cout << "\t iTDRL(r,";
						}
						for(it=rhos2[k-1][0].begin(); it!=rhos2[k-1][0].end(); ++it){
							if(*it<0){
								cout << " -" << nmap[abs(*it)];
							}else{
								cout <<  " " << nmap[*it];
							}
						}
						cout << ",";
						for(it=rhos2[k-1][1].begin(); it!=rhos2[k-1][1].end(); ++it){
							if(*it<0){
								cout << " -" << nmap[abs(*it)];
							}else{
								cout <<  " " << nmap[*it];
							}
						}
						cout << ")" << endl;
					}
				}
			}
		}
	}
	return 0;
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void getoptions(int argc, char *argv[], string &fname, bool &print_distance){
	int c;
	while (1) {
		static struct option long_options[] = {
			{"file",    required_argument, 0, 'f'},
			{"distance", no_argument, 0, 'd'},
			{"help", no_argument, 0, 'h'},
			{0, 0, 0, 0}
		};
        int option_index = 0;			// getopt_long stores the option index here.
        c = getopt_long (argc, argv, "f:hd", long_options, &option_index);
        /* Detect the end of the options. */
        if (c == -1)
        	break;
        switch (c){
			case 0: // If this option set a flag, do nothing else now.
				if (long_options[option_index].flag != 0)
					break;
				cout << "option "<< long_options[option_index].name;
				if (optarg)
					cout << " with arg " << optarg;
				cout << endl;
				break;
			case 'f':
				fname = optarg;
				break;
			case 'd':
				print_distance=true;
				break;
			case 'h':
				usage();
				exit(EXIT_SUCCESS);
			case '?':
				usage();
				exit(EXIT_SUCCESS);
				break; /* getopt_long already printed an error message. */
			default:
				usage();
        }
	}
}

// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

void usage(){
	cerr << "Tool solves the sorting problem and the distance problem for two given permutations under the inverse tandem duplication random loss (iTDRL) rearrangement model. "<<endl;
	cerr << " --file		-f	<fasta file>, contains (multiple gene orders)" << endl;
	cerr << " --distance	-d	print only a table with iTDRL distances" << endl;
	cerr << " --help		-h	prints this message" << endl;
	cerr << "required argument: f" << endl;
	exit(1);
}
