#include <algorithm>
#include <limits>
#include <math.h>
#include <iostream>
#include <vector>
#include <string>
#include <sstream>

#include "helpers.hpp"

using namespace std;

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* STRING FUNCTIONS  ++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void split_string(const string& str,
                      vector<string>& tokens,
                      const string& delimiters){
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
    }
}

///////////////////////////////////////////////////////////////////////////////

void strip_leading_ws(string &str){
    str.erase(str.begin(), std::find_if(str.begin(), str.end(), std::not1(std::ptr_fun<int, int>(std::isspace))));
}

///////////////////////////////////////////////////////////////////////////////

void strip_trailing_ws(string &str){
	str.erase(std::find_if(str.rbegin(), str.rend(), std::not1(std::ptr_fun<int, int>(std::isspace))).base(), str.end());
}

/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* iTDRL FUNCTIONS ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void decomposition(genome g, vector< set<int> > &deco){
	set<int> temp;
	for(unsigned int i=0; i<g.size()-1; i++){
		// i is end of maximum increasing string
		if(g[i]>g[i+1]){
			temp.insert(g[i]);
			deco.push_back(temp);
			temp.clear();
		}else{ // inside of max increasing substring
			temp.insert(g[i]);
		}
	}
	// last element:
	temp.insert(g[g.size()-1]);
	deco.push_back(temp);
}

///////////////////////////////////////////////////////////////////////////////

unsigned int dist_iTDRL(genome g){
	unsigned int dist=0;
	unsigned int subs=max_inc_substrings(g);
	if(subs<=0){
		cerr << "ERROR: desc(genome g) function returns 0, when using: " << g << "." << endl;
		exit(EXIT_FAILURE);
	}
	//distance formular
	vector<int> iota;
	for(unsigned int i=1; i<=g.size(); i++){
		iota.push_back(i);
	}
	if(g.getChromosom()==iota){
		return dist;
	}else{
		dist=ceil( log2((float)subs) ) + 1;
		if(((subs==2) ||(subs % 2 == 1)) && (g.chromosom[0]>0) && (g.chromosom[g.size()-1]<0)){
			for(unsigned int k=0; pow(2,k)+1 <= subs; k++){
				if(subs==pow(2,k)+1){
					dist=log2((float)subs-1)+1;
				}
			}
		}
		return dist;
	}
}

///////////////////////////////////////////////////////////////////////////////

unsigned int max_inc_substrings(genome g){
	int count=1;
	for(unsigned int i=0; i<g.size()-1; i++){
		// i is end of maximum increasing string
		if(g[i]>g[i+1]){
			count+=1;
		}
	}
	return count;
}

///////////////////////////////////////////////////////////////////////////////

genome multiply(genome pi, genome sigma){
	if(pi.size() != sigma.size()){
		cerr << "ERROR: pi and sigma do not have the same size! ---> exiting!" << endl;
		exit(EXIT_FAILURE);
	}
	genome product;
	product.setCircular(sigma.getCircular());
	product.set_nmap(sigma.get_nmap());

	//π°σ is the function that maps any element x of the set to π(σ(x))
	vector<int> tmp;
	for(unsigned int x=0; x< pi.size(); x++){
		int value;
		value=abs(pi[abs(sigma[x])-1]);
		// take care about the signs
		if((sigma[x]<0 && pi[abs(sigma[x])-1]>0) || (sigma[x]>0 && pi[abs(sigma[x])-1]<0)){
			value*=-1;
		}
		tmp.push_back( value );
	}
	product.setChromosom(tmp);
	return product;
}

///////////////////////////////////////////////////////////////////////////////

void sort_by_iTDRL(genome g, vector<genome> &scen, vector<vector<set<int>>> &iTDRLs, bool &last){
	// check g != iota
	vector<int> iota;
	last=false;
	for(unsigned int i=1; i<=g.size(); i++){
		iota.push_back(i);
	}
	if(g.getChromosom()!=iota){
		genome running_g=g;
		unsigned int k=dist_iTDRL(g);
		for(unsigned int i=k; i>0; i--){
			vector<set<int>> temp_iTDRL;
			genome temp_genome;
			//cout << "running: "  << running_g << endl;
			if((i==1) && (running_g[0]<0)){
				//cout << "CASE 1" << endl;
				sort_case_1(running_g, temp_genome, temp_iTDRL);
				last=true;
			}else{
				//cout << "CASE 2" << endl;
				sort_case_2(running_g, temp_genome, temp_iTDRL);
			}
			running_g=temp_genome;
			scen.push_back(temp_genome);
			iTDRLs.push_back(temp_iTDRL);
		}
	}//g != iota
}

///////////////////////////////////////////////////////////////////////////////

void sort_by_iTDRL(genome source, genome target, vector<genome> &scen, vector<vector<set<int>>> &iTDRLs, bool &last){
	if(source.size()!=target.size()){
		cerr << "ERROR: genomes do not have the same length: " << source << " <-> " << target << "! ---> exiting!" << endl;
		exit(EXIT_FAILURE);
	}

	//rename to original gene names
	vector<int > tmp = source.inverse(true);
	genome inverse;
	inverse.setChromosom(tmp);
	inverse.setCircular(source.getCircular());
	inverse.set_nmap(source.get_nmap());

	// pi -> sigma ~ iota=\pi^{-1}\circ\pi -> \pi^{-1}\circ\sigma
	// \pi^{-1}\circ\sigma
	genome temp=multiply(inverse, target);

	scen.push_back(temp);
	sort_by_iTDRL(temp, scen, iTDRLs, last);


	for(unsigned int i=0; i<scen.size(); i++){
		if(i==scen.size()-1){
			genome bla=multiply(scen[i],source);
			scen[i]=bla;
		}else if(i==0){
			genome bla=multiply(source,scen[i]);
			scen[i]=bla;
		}else{
			genome bla=multiply(source,scen[i]);
			scen[i]=bla;
		}
	}

	//rename the genes of the rrrmts
	set<int>::iterator it;

	for(unsigned int i=0; i<iTDRLs.size();i++){
		vector<set<int>> temp;
		set<int> L;
		set<int> R;
		if(i==iTDRLs.size()-1 && last){
			for(it=iTDRLs[i][0].begin(); it!=iTDRLs[i][0].end(); ++it){
				L.insert((-1)*scen[i][*it]);
			}
			for(it=iTDRLs[i][1].begin(); it!=iTDRLs[i][1].end(); ++it){
				R.insert((scen[i][*it]));
			}
		}else{
			for(it=iTDRLs[i][0].begin(); it!=iTDRLs[i][0].end(); ++it){
				L.insert(scen[i][*it]);
			}
			for(it=iTDRLs[i][1].begin(); it!=iTDRLs[i][1].end(); ++it){
				R.insert((-1)*(scen[i][*it]));
			}
		}
		temp.push_back(L);
		temp.push_back(R);
		iTDRLs[i]=temp;
	}
}
///////////////////////////////////////////////////////////////////////////////

void sort_case_1(genome g, genome &sigma, vector<set<int>> &iTDRLs){
	vector< set<int> > deco;
	set<int>::iterator it;
	// get decomposition
	decomposition(g,deco);
	vector<int> chromosome;
	if(deco.size()>1){
		cerr << "ERROR: temp_genome: " << g << " does have more than 1 maximal increasing substrings! ---> exiting!" << endl;
		exit(EXIT_FAILURE);
	}

	// split the single maximum increasing run up in two sets
	set<int> minus;
	set<int> plus;
	set<int> result;
	for(it=deco[0].begin(); it!=deco[0].end(); ++it){
		if((*it)<0){
			minus.insert(*it);
		}else{
			plus.insert(*it);
		}
	}
	oplus(result,plus,minus);
	for(it=result.begin(); it!=result.end(); ++it){
		chromosome.push_back(*it);
	}

	set<int> same_sign;
	set<int> diff_sign;
	// get positions of the elements that are in L and in R
	for(unsigned int pos=0; pos<g.size();pos++){
		if(g.chromosom[pos] < 0){
			diff_sign.insert((int)pos);
		}else{
			same_sign.insert((int)pos);
		}
	}

	iTDRLs.push_back(diff_sign);
	iTDRLs.push_back(same_sign);

	sigma.setChromosom(chromosome);
	sigma.setCircular(g.getCircular());
	sigma.set_nmap(g.get_nmap());

}

///////////////////////////////////////////////////////////////////////////////

void sort_case_2(genome g, genome &sigma, vector<set<int>> &iTDRLs){
	set<int> same_sign;
	set<int> diff_sign;
	vector< set<int> > deco;
	set<int>::iterator it;
	// get decomposition
	decomposition(g,deco);
	vector<int> chromosome;

	if(deco.size()%2==0){//l is even
		for(unsigned int i=0; i<deco.size()/2;i++){
			set<int> result;
			oplus(result,deco[i],deco[deco.size()-i-1]);
			for(it=result.begin(); it!=result.end(); ++it){
				chromosome.push_back(*it);
			}
		}
			//get positions of the elements that are in L and in R
			unsigned int bp=0;
			for(unsigned int ite=0; ite<(float)deco.size()/2; ite++){
				bp+=deco[ite].size();
			}
			for(unsigned int pos=0; pos<g.size();pos++){
				if(pos < bp){
					same_sign.insert((int)pos);
				}else{
					diff_sign.insert((int)pos);
				}
			}
		}else{// l is odd
			for(unsigned int i=0; i<ceil((float)deco.size()/2)-1;i++){
				set<int> result;
				oplus(result,deco[i],deco[deco.size()-i-1]);
				for(it=result.begin(); it!=result.end(); ++it){
					chromosome.push_back(*it);
				}
			}
			set<int> minus;
			set<int> plus;
			set<int> result;
			for(it=deco[ceil((float)deco.size()/2)-1].begin(); it!=deco[ceil((float)deco.size()/2)-1].end(); ++it){
				if((*it)<0){
					minus.insert(*it);
				}else{
					plus.insert(*it);
				}
			}
			oplus(result,minus,plus);
			for(it=result.begin(); it!=result.end(); ++it){
				chromosome.push_back(*it);
			}
			//get positions of the elements that are in L and in R
			unsigned int bp=0;
			for(unsigned int ite=0; ite<ceil((float)deco.size()/2)-1; ite++){
				bp+=deco[ite].size();
			}
			bp+=minus.size();
			for(unsigned int pos=0; pos<g.size();pos++){
				if(pos < bp){
					same_sign.insert((int)pos);
				}else{
					diff_sign.insert((int)pos);
				}
			}
		}

		iTDRLs.push_back(same_sign);
		iTDRLs.push_back(diff_sign);

		sigma.setChromosom(chromosome);
		sigma.setCircular(g.getCircular());
		sigma.set_nmap(g.get_nmap());
}

///////////////////////////////////////////////////////////////////////////////

void test_sorting_algorithm(unsigned int size, unsigned int numb){
	genome g(size, false, false);
	genome iota(size, false, false);
	srand (time(NULL));
	//srand (43);

	vector<string> nmap;
	nmap.push_back("INVALID_GENE_0");
	for(unsigned int i=1; i< size+1; i++){
		stringstream ss;
		ss << i;
		nmap.push_back(ss.str());
	}
	g.set_nmap(&nmap);
	iota.set_nmap(&nmap);
	for( unsigned r = 0; r<numb; r++ ){
		std::cout << "progress: " << ((float)r*100)/(float)numb << "%" << "\r" << std::flush;
		// build permutation
		vector<int> temp_g=g.getChromosom();
		// build permutation by fisher-yates-shuffle
		for(int i = 0; i < (int)temp_g.size(); i++) {
			int j =  rand() % (temp_g.size()-i)  + i;
			int temp = temp_g[i];
			temp_g[i] = temp_g[j];
			temp_g[j] = temp;
		}
		// assign sign to every element
		for(int i=0; i< (int)temp_g.size(); ++i){
			int j= rand() % 2;
			if(j==0){
				temp_g[i]*=-1;
			}
		}
		g.setChromosom(temp_g);

		// sort permutation
		set<int>::iterator it;
		vector<genome> sigmas;
		vector<vector<set<int>>> rhos;
		bool last;
		sort_by_iTDRL(iota, g, sigmas, rhos, last);

		// check for correct solution
		// transform g with the scenario into iota and check if the transformation is correct
		genome running_genome=iota;
		for( int k=sigmas.size()-1; k>=0; --k){
			if(k!=0){
				//check whether liTDRL or riTDRL
				bool liTDRL=false;
				if(k==(int)sigmas.size()-1 && last){
					liTDRL=true;
				}
				// L==0; R==1
				vector<bool> TDRL(running_genome.size(),false);
				for(it=rhos[k-1][0].begin(); it!=rhos[k-1][0].end(); ++it){
					 TDRL[abs(*it)-1]=false;
				}
				for(it=rhos[k-1][1].begin(); it!=rhos[k-1][1].end(); ++it){
					 TDRL[abs(*it)-1]=true;
				}
				vector<int> L;
				vector<int> R;
				if(liTDRL){
					for(unsigned int z=0; z<running_genome.size(); z++){
						if(TDRL[abs(running_genome[z])-1]==false){
							L.push_back((-1)*running_genome[z]);
						}else{
							R.push_back(running_genome[z]);
						}
					}
					reverse(L.begin(),L.end());
				}else{
					//riTDRL
					for(unsigned int z=0; z<running_genome.size(); z++){
						if(TDRL[abs(running_genome[z])-1]==false){
							L.push_back(running_genome[z]);
						}else{
							R.push_back((-1)*running_genome[z]);
						}
					}
					reverse(R.begin(),R.end());
				}
				// build new genome
				vector<int> temp_genome;
				// add L and R
				for(unsigned int z=0; z<L.size();z++){
					temp_genome.push_back(L[z]);
				}
				for(unsigned int z=0; z<R.size();z++){
					temp_genome.push_back(R[z]);
				}
				// update iota_copy
				running_genome.setChromosom(temp_genome);
			}else{
				if(running_genome.getChromosom()!=g.getChromosom()){
					cerr << "ERROR: scenario seems to contain an error! ---> exiting!" << endl;
					cerr << "Scenario:" << sigmas[sigmas.size()-1] << " to " << sigmas[0] << endl;
					for(unsigned int i=0; i< running_genome.size(); i++){
						cerr << running_genome[i] << " -- "<< g[i] << endl;
					}
					 for(int z=sigmas.size()-1; z>=0; --z){
						 cerr << sigmas[z] << endl;
						if(z!=0){
							if(z==(int)sigmas.size()-1 && last){
								cerr << "\t iTDRL(l,";
							}else{
								cerr << "\t iTDRL(r,";
							}
							for(it=rhos[z-1][0].begin(); it!=rhos[z-1][0].end(); ++it){
								if(*it<0){
									cout << " -" << nmap[abs(*it)];
								}else{
									cout <<  " " << nmap[*it];
								}
							}
							cerr << ",";
							for(it=rhos[z-1][1].begin(); it!=rhos[z-1][1].end(); ++it){
								if(*it<0){
									cout << " -" << nmap[abs(*it)];
								}else{
									cout <<  " " << nmap[*it];
								}
							}
							cerr << ")" << endl;
						}
					}
					cerr << endl;
					exit(EXIT_FAILURE);
				}
			}
		}
	}
	cerr << "success"<<endl;
}

///////////////////////////////////////////////////////////////////////////////

void oplus(set<int> &result, set<int> first, set<int> second){
	set<int>::iterator it;
	for(it=first.begin(); it!=first.end(); ++it){
		result.insert(*it);
	}
	for(it=second.begin(); it!=second.end(); ++it){
		result.insert((-1)*(*it));
	}
}

///////////////////////////////////////////////////////////////////////////////
