/**
 * @file genome.cpp
 * genom functions
 * @author T. Hartmann and M. Bernt
 */

#include "genome.hpp"

#include <algorithm>
#include <iostream>
#include <iterator>
#include <limits>
#include <map>

using namespace std;

///////////////////////////////////////////////////////////////////////////////
genome::genome(){
	//~ name = "";
	chromosom.clear();
	circular = false;
	nmap = NULL;
	return ;
}

///////////////////////////////////////////////////////////////////////////////

genome::genome(int cnt, char c, bool inv) {
	chromosom.clear();
	chromosom = vector<int>(cnt, 0);
	for (int i = 1; i <= cnt; i++) {
		chromosom[i-1] = i;
	}
	nmap = NULL;
	circular = c;

	if( inv ){
		reverse (chromosom.begin(), chromosom.end());
		for(unsigned i=0; i<chromosom.size(); i++){
			chromosom[i] *= -1;
		}
	}
}

///////////////////////////////////////////////////////////////////////////////

genome::genome(const vector<int> &p, int c, vector<string> *nm, int norm_to){
	chromosom = p;
	circular = c;
	if( c ){
		normalize(norm_to);
	}
	nmap = nm;
}

///////////////////////////////////////////////////////////////////////////////

void genome::de_identify_g( const genome &origin ){
	genome tmp( size(), 0 ); 	// temporary genom

	for( unsigned i=0; i<size(); i++ ){
		tmp[ i ] = origin[ abs(chromosom[i])-1 ];
		if( chromosom[i] < 0)
			tmp[ i ] *= -1;
	}

	chromosom = tmp.chromosom;
}

///////////////////////////////////////////////////////////////////////////////

vector<string> *genome::get_nmap() const{
	return nmap;
}

///////////////////////////////////////////////////////////////////////////////

vector<int> genome::getChromosom() const{
	return chromosom;
}

///////////////////////////////////////////////////////////////////////////////

char genome::getCircular() const{
	return circular;
}

///////////////////////////////////////////////////////////////////////////////

genome genome::identify_g(const genome &q) const{
	unsigned s = size();
	static genome g;
	static vector<int> pi_inv;

	if(g.size() != s){
		g.chromosom.resize(s);
		pi_inv.resize(s);
	}

		// get the inverse permutation of p [this] -> p^-1
		// this is the permutation where the i-th element
		// is the position of the i-th element in p
		// (+ possible sign swaps)
	for(unsigned i=0; i<s; i++){
		pi_inv[ abs(chromosom[i]) - 1 ] = i+1;
		if(chromosom[i]<0)
			pi_inv[abs(chromosom[i]) - 1] *= -1;
	}

		// compute (p^-1) * q
		// this is the permutation where the i-th element is the
		// element of p^-1 at the position of the i-th element of q
	for(unsigned i=0; i<s;i++){
		if(q[i]<0)
			g[i] = -1 * pi_inv[abs(q[i]) - 1];
		else
			g[i] = pi_inv[abs(q[i]) - 1];
	}
	return g;
}

///////////////////////////////////////////////////////////////////////////////

vector<int> genome::inverse( bool sign ) const{


	unsigned s=size();
	vector<int> inv(s, std::numeric_limits< int >::max());	// the inverse permutation

	for(unsigned i=0; i<chromosom.size(); i++){
		inv[ abs(chromosom[i])-1 ] = i+1;
		if( sign && chromosom[i] < 0){
			inv[ abs(chromosom[i])-1 ] *= -1;
		}
	}
	return inv;
}

///////////////////////////////////////////////////////////////////////////////

void genome::normalize(int to){
	unsigned onePos = 0;

		// search element to
	for(unsigned i=0; i<chromosom.size(); i++){
		if(abs(chromosom[i]) == abs(to)){
			onePos = i;
			break;
		}
	}

		// if its in the wrong direction -> rotate
	if(chromosom[onePos] == -1 * to){
		std::reverse(chromosom.begin(), chromosom.end());
		for(unsigned i=0; i<chromosom.size(); i++)
			chromosom[i] *= -1;
	}

		// bring the element to the front
	rotate(chromosom.begin(), find(chromosom.begin(), chromosom.end(), to), chromosom.end());
}

///////////////////////////////////////////////////////////////////////////////

void genome::push_back(int gene) {
	chromosom.push_back(gene);
}

///////////////////////////////////////////////////////////////////////////////

void genome::setChromosom(vector<int> c) {
	chromosom = c;
}

///////////////////////////////////////////////////////////////////////////////

void genome::setCircular(char c) {
	circular = c;
	if(circular)
		normalize();
}

///////////////////////////////////////////////////////////////////////////////
void genome::set_nmap( vector<string> *nm ){
	nmap = nm;
}

///////////////////////////////////////////////////////////////////////////////

unsigned int genome::size() const{
	return chromosom.size();
}

///////////////////////////////////////////////////////////////////////////////

int& genome::operator[](unsigned i){
	return chromosom[i];
}

///////////////////////////////////////////////////////////////////////////////

int genome::operator[](unsigned i) const{
	return chromosom[i];
}

///////////////////////////////////////////////////////////////////////////////

bool genome::operator==(const genome& g) const {
	return equal(chromosom.begin(), chromosom.end(), g.chromosom.begin());
}

///////////////////////////////////////////////////////////////////////////////

bool genome::operator!=(const genome& g) const {
	if (*this == g)
		return false;
	else
		return true;
}

///////////////////////////////////////////////////////////////////////////////

bool genome::operator<(const genome& g) const {

	unsigned s = min(this->size(), g.size());
	for (unsigned i = 0; i < s; i++) {

		if ( chromosom[i] > g.chromosom[i]) {
			return false;
		} else if (chromosom[i] < g.chromosom[i]){
			return true;
		}
	}

	// -> up to s both are equal
	// -> decide by length

	if( chromosom.size() < g.size() )
		return true;
	else
		return false;
}


///////////////////////////////////////////////////////////////////////////////

ostream &operator<<(ostream &out, const genome &g) {
	for (unsigned i = 0; i < g.size(); i++) {
		if( g.nmap == NULL ){
			out.width(3);
			out << g.chromosom[i];
		}else{
			if( g.chromosom[i] < 0 ){
				out << "-";
			}else{
				out<<"";
			}
			out << (*g.nmap)[ abs(g.chromosom[i]) ];
		}

		if( i < g.size() -1  )
			out << " ";
	}
	return out;
}

///////////////////////////////////////////////////////////////////////////////

ostream & operator<<(ostream &out, const vector<genome> &genomes){
	for (unsigned i=0; i<genomes.size(); i++){
		out << genomes[i] <<endl;
	}

	return out;
}

///////////////////////////////////////////////////////////////////////////////

