/**
 * @file genome.hpp
 * genome header file
 * @author Tom Hartmann and M. Bernt
 */

#ifndef _GENOME_HPP_
#define _GENOME_HPP_

#include <iterator>
#include <map>
#include <string>
#include <set>
#include <vector>
#include <stdexcept>
#include <iostream>

using namespace std;

class genome;


/**
 * class for a genom.
 */
class genome{
	private:
		vector<string> *nmap;
		/*! chromosom of the genom*/
	public:
		vector<int> chromosom;
		/*! flag thats true if the genome is circular */
		char circular;

		/**
		 * constructs an empty new genome
		 */
		genome();

		/**
		 * constructs a new idendity genom with length cnt
		 *@param cnt length of the genom
		 *@param c circularity of the genom
		 *@param[in] inv true -> construct inverse of the id
		 */
		genome(int cnt, char c, bool inv=false);

		/**
		 * contructs a new genome with the permutation given in an int vector
		 * @param[in] p permuation
		 * @param[in] c circularity of the genom
		 * @param[in] nm pointer to a name map
		 * @param[in] norm_to normalize circular genomes to this element (default 1)
		 */
		genome(const vector<int> &p, int c, vector<string> *nm = NULL, int norm_to = 1);

		/**
		 * reverse the effect of origin.identify_g(goal)
		 * @param[in] origin the origin genom used in identify
		 */
		void de_identify_g( const genome &origin );

		/**
		 * return the pointer to the namemap
		 * @return the pointer
		 */
		vector<string> *get_nmap() const;

		/**
		 * getter for the private member circular
		 *@return circular
		 */
		char getCircular() const;

		/**
		 * getter for the chromosom
		 *@return the chromosom vector
		 */
		vector<int> getChromosom() const;

		/**
		 * computes the permutation p = p(q) (where p is this)
		 *@param[in] q the 2nd genome
		 *@return the 'identified' genome
		 */
		genome identify_g(const genome &q) const;

		/**
		 * get the inverse permutation of a signed permutation
		 * (index i of the inverse stores the position of the element i
		 * in permutation g, if element i is negative then the element of
		 * the inverse is also negative)
		 * @param[in] sign add signs to the inverse (default: true)
		 * @return the inverse permutation
		 */
		vector<int> inverse( bool sign = true) const;

		/**
		 * normalizes a genome:
		 * - if circular: rotate such that the element to is at position 0
		 *@param[in] to normalize to this element (default 1)
		 */
		void normalize(int to=1);

		/**
		 * append a new gene to the end of the chromosom vector
		 * @param gene the gene
		 */
		void push_back(int gene);

		/**
		 * setter for the chromosom
		 *@param c the new chromosom vector
		 */
		void setChromosom(vector<int> c);

		/**
		 * set the namemap
		 * @param[in] nm the new name map
		 */
		void set_nmap( vector<string> *nm );

		/**
		 * setter for the circular property
		 * the normalize function is also called
		 *@param c the circularity of the genome
		 */
		void setCircular(char c);

		/**
		 * getter for the size (lenght) of the chromosom
		 *@return the size
		 */
		unsigned int size() const;

		/**
		 * return a reference to the element element at index i
		 * of the chromosom. This can be used to get the element
		 * or write it.
		 *@param[in] i the index
		 *@return the element at index i
		 */
		int& operator[](unsigned i);

		/**
		 * return the element element at index i of the chromosom.
		 * This can be used to get the element.
		 *@param[in] i the index
		 *@return the element at index i
		 */
		int operator[](unsigned i) const;

		/**
		 * checks if the chromosomes of the calling genom and the genom g are equal
		 *@param g genom
		 *@return true if equal false else
		 */
		bool operator==(const genome& g) const;

		/**
		 * checks if the chromosomes of the calling genom and the genom g are unequal
		 *@param g genom
		 *@return true if unequal false else
		 */
		bool operator!=(const genome& g) const;

		/**
		 * checks if the chromosomes are smaller (needed for saving in a map)
		 * the test is like a string < test; it's important that the genomes have
		 * equal length!!
		 *@param g genom
		 *@return true if < false else
		 */
		bool operator<(const genome& g) const;

		/**
		 * output operator for the genom (outputs the genom and then the name)
		 *@param out the stream to output in
		 *@param g the genom to output
		 *@return output
		 */
		friend ostream & operator<<(ostream &out, const genome &g);

		/**
		 * output operator for a vector of genomes
		 * it prints the Genomes and the distancematrix
		 *@param out the stream to output in
		 *@param G the genomes to output
		 *@return output
		 */
		friend ostream & operator<<(ostream &out, const vector<genome> &genomes);

};


#endif//_GENOME_HPP_
